
class Buttons():

    ANY    = -1
    A      =  0
    B      =  1
    X      =  2
    Y      =  3
    LB     =  4
    RB     =  5
    VIEW   =  6
    MENU   =  7
    LSTICK =  8
    RSTICK =  9

    MOUSE_LEFT   = 1
    MOUSE_CENTER = 2
    MOUSE_RIGHT  = 4


class Gamepads():

    ANY = -1

class Axis():

    ANY = -1
    LX = "x"
    LY = "y"
    RX = "rx"
    RY = "ry"
    Z  = "z"

